-- This work is licensed under a Creative Commons
-- Attribution-NonCommercial-ShareAlike 4.0 International License.
-- http://creativecommons.org/licenses/by-nc-sa/4.0

-- SpoC-128_sLiSCP-light-[256]

-- 2022 HW design:  Mark D. Aagaard, Nusa Zidaric
-- 2022 VHDL implementation: Nusa Zidaric
-- note overlap with ACE modules at
-- ( http://uwaterloo.ca/communications-security-lab/lwc/ace )

-- SpoC submitters/designers:
-- Riham AlTawy
-- Guang Gong
-- Morgan He
-- Ashwin Jha
-- Kalikinkar Mandal
-- Mridul Nandi
-- Raghvendra Rohit

-- http://uwaterloo.ca/communications-security-lab/lwc/spoc

-- This is a human-readable summary of (and not a substitute for) the license.
-- You are free to:

--     Share:  copy and redistribute the material in any medium or format
--     Adapt:  remix, transform, and build upon the material

--     The licensor cannot revoke these freedoms as long as you follow
--     the license terms.

-- Under the following terms:

--     Attribution - You must give appropriate credit, provide a link to
--     the license, and indicate if changes were made. You may do so in
--     any reasonable manner, but not in any way that suggests the
--     licensor endorses you or your use.

--     NonCommercial - You may not use the material for commercial
--     purposes.

--     ShareAlike - If you remix, transform, or build upon the material,
--     you must distribute your contributions under the same license as
--     the original.

--     No additional restrictions - You may not apply legal terms or
--     technological measures that legally restrict others from doing
--     anything the license permits.
---------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.spoc_128_pkg.all;
use work.util.all;
use work.ctrl_pkg.all;

entity ctrl is
  port
    ( clk
    , reset          : in  std_logic
    ; pre_mode       : in  mode_ty
    ; pre_ds_ctrl    : in  ds_ctrl_ty
    ; pre_valid      : in  std_logic
    ; post_ready     : in  std_logic
    ; pre_padding    : in  pad_ty
    ; o_control     	: out dp_op_ty
    ; o_k             : out std_logic
    ; o_new_ctrl 	  	: out en_ty
    ; o_pad_ctrl	  	: out padxor_ty
    ; post_valid     : out std_logic
    ; pre_ready      : out std_logic

   );
end entity;

architecture rtl of ctrl is

  signal state, state_next	: state_ty;
  signal count  					  : count_ty;
  signal k_count, t_count  	: std_logic;

  signal new_ctrl 	 : en_ty;
  signal pad_ctrl 	 : padxor_ty;
  signal control 		 : dp_op_ty ;


  signal exit_l_load, exit_perm
      ,  exit_m_main, exit_t_tag  : 		std_logic;

  signal condition_next_l_load
      ,  condition_next_i_main, condition_next_m_main
      ,  condition_next_t_tag , condition_next_perm
       	: std_logic;

  signal rollover, counting, flip,  newbit:					std_logic;

  signal r_load, r_main, r_ed, r_ad
      ,  q_load, q_main, q_ed, q_tag
			,	 ds_ctrl0, ds_ctrl1, ds_ctrl2
      ,  ds_tag, ds_perm
      ,  v_ed, v_tag:  						std_logic;

  begin

  ------------------------------------------------------------
  -- environment control
  ------------------------------------------------------------
  pre_ready 	<= '1' when x2b( r_load or r_main or r_ad or r_ed ) else	'0';
  post_valid 	<= '1' when x2b( v_ed or v_tag) 	else	'0';

  ------------------------------------------------------------
  -- datapath control
  ------------------------------------------------------------
  o_k <= k_count;

  o_control(en_load_idx) 			 <= 	control(en_load_idx) ;
  o_control(sel_load_perm_idx) <= 	control(sel_load_perm_idx);
  o_control(sel_round_step_idx) <= 	control(sel_round_step_idx) and
                                    count(0) and count(1) and count(2);

  o_control(lfsr_c_en_idx)			<= (control(lfsr_c_en_idx) and is_perm_st (state))
                                    or ds_tag;
  o_control(lfsr_c_reset_idx) 	<= 	control(lfsr_c_reset_idx) or reset;

  o_control(sel_ds_idx)     <= 	control(sel_ds_idx);
  o_control(sel_input0_idx) <= 	control(sel_input0_idx);
  o_control(sel_input1_idx) <= 	control(sel_input1_idx);
  o_control(sel_out_idx)    <= 	control(sel_out_idx);
  o_control(en_out_idx)     <= 	control(en_out_idx) and ( v_ed or v_tag );

  u_ce : entity work.ctrl_pad_to_en port map (pre_padding, new_ctrl, pad_ctrl);

  o_new_ctrl 	<= new_ctrl 	when  x2b( v_ed and is_dec_mode(pre_mode)   )  else
  						(others => newbit);
  o_pad_ctrl 	<= pad_ctrl when  x2b( v_ed and is_dec_mode(pre_mode)   )  else
  						(others => '0');
  newbit      <= '1' when x2b( v_ed or r_ad ) else '0';

  control <=
      dp_op_1 when x2b( q_load and pre_valid)  								              else
  	  dp_op_2 when x2b( is_perm_st(state) and not is_perm_done(count))			else
  		dp_op_3	when x2b(	r_ad  )					                                    else
  		dp_op_4	when x2b( v_ed and is_enc_mode(pre_mode) )   	 	              else
 			dp_op_5 when x2b( v_ed and is_dec_mode(pre_mode) )  	                else
  		dp_op_6 when x2b( ds_tag)					                                    else
			dp_op_7 when x2b( v_tag ) 															              else
			(others => '0');

  ------------------------------------------------------------
  -- transient states
  ------------------------------------------------------------

	  exit_l_load   <= is_l_load_st(state)  and ( not k_count );
	  exit_perm   	<= is_perm_st(state)    and is_perm_done( count );
	 	exit_t_tag  	<= is_t_tag_st(state)   and post_ready;
	 	exit_m_main  	<= is_m_main_st(state)  and ( not t_count );

    r_load <= is_i_load_st( state ) or								-- from idle
            ( is_l_load_st( state ) and   k_count );  -- loop back
    q_load <= r_load;

    r_main <= exit_l_load                  or         -- first in
              is_i_main_st( state ) 			 or 				-- from idle/stall
            ( exit_perm and (not t_count)) or					-- loop back   (from perm)
            ( exit_m_main)  						;							-- loop back   (from main)

    q_main <= ( r_main )   ;                  				-- first in

    -- q_main <= is_i_main_st( state ) 			 or 				-- from idle/stall
    --       			  		  ( r_main )   ;                  				-- first in

    ds_ctrl0 <= q_main   and     pre_valid;
    ds_tag   <= ds_ctrl0 and     is_tag_ds(pre_ds_ctrl);
    ds_ctrl1 <= ds_ctrl0 and not is_tag_ds(pre_ds_ctrl);
    q_ed  	 <= ds_ctrl1 and     is_ed_ds(pre_ds_ctrl);
    ds_ctrl2 <= ds_ctrl1 and not is_ed_ds(pre_ds_ctrl);
    r_ad  	 <= ds_ctrl2 and     is_ad_ds(pre_ds_ctrl);
    ds_perm  <= ds_ctrl2 and not is_ad_ds(pre_ds_ctrl);

    r_ed  <= q_ed and post_ready;
    v_ed 	<= r_ed;

    v_tag <= is_t_tag_st( state );
    q_tag <= v_tag;

  ------------------------------------------------------------
  -- next state logic
  ------------------------------------------------------------

    counting <= 	'1' when x2b( is_perm_st(state) and not is_t_tag_st(state))
  	              else '0';
    rollover <=  r_main;
    flip 	   <=  (q_load and pre_valid );

    condition_next_l_load	<=   q_load and pre_valid; 				  -- wait for parcel
    condition_next_i_main	<= ( q_main and not pre_valid )  or -- wait for parcel
    									       ( q_ed   and not post_ready ) ;	-- stall

    condition_next_m_main <=   ds_tag or v_ed or r_ad ;				-- main mode

	  condition_next_perm  	 <=
          ds_perm                               or					-- start perm in main
	 			( is_m_main_st(state)  and   t_count ) 	or		        -- start tag perm
	  		( is_perm_st(state) and not is_perm_done( count )) ; 	-- continue perm

    condition_next_t_tag  <= ( exit_perm and t_count ) or   -- first in
                             ( q_tag and not post_ready );  -- stall

    state_next <= l_load_st  when x2b( condition_next_l_load  ) else
       						i_main_st  when x2b( condition_next_i_main  )	else
       						m_main_st  when x2b( condition_next_m_main  )	else
       						perm_st    when x2b( condition_next_perm    )	else
       					  t_tag_st   when x2b( condition_next_t_tag   )	else
       			 	    i_load_st;

  ------------------------------------------------------------
  -- state update
  ------------------------------------------------------------

  state_update: process(clk) begin
    if rising_edge(clk) then
      if x2b( reset ) then
        state  	<= i_load_st;
       else
        state  	<= state_next;
       end if;
    end if;
  end process;

  count_update: process(clk) begin
    if rising_edge(clk) then
       if x2b( reset ) then
       	count  	<= ( others => '0');
       else
        if 	x2b(rollover) then
          count  	<= ( others => '0');
       	elsif x2b(counting) then
      		count <= count + 1;
        	end if;
       end if;
    end if;
  end process;

  k_update: process(clk) begin
    if rising_edge(clk) then
      if x2b( reset ) then
        k_count <= '0';
      elsif x2b(flip) then
        k_count <= not k_count;
      end if;
    end if;
  end process;

  t_update: process(clk) begin --merge with k_count if enough time
    if rising_edge(clk) then
      if x2b( reset ) then
        t_count <= '0';
      elsif x2b(ds_tag) then
        t_count <= '1';
      elsif x2b(q_tag and post_ready) then
        t_count <= '0';
      end if;
    end if;
  end process;

end architecture;
