-- This work is licensed under a Creative Commons
-- Attribution-NonCommercial-ShareAlike 4.0 International License.
-- http://creativecommons.org/licenses/by-nc-sa/4.0

-- 2022 HW design:  Mark D. Aagaard, Nusa Zidaric
-- 2022 VHDL implementation: Nusa Zidaric
-- note overlap with ACE modules at
-- ( http://uwaterloo.ca/communications-security-lab/lwc/ace )

-- SpoC submitters/designers:
-- Riham AlTawy
-- Guang Gong
-- Morgan He
-- Ashwin Jha
-- Kalikinkar Mandal
-- Mridul Nandi
-- Raghvendra Rohit

-- http://uwaterloo.ca/communications-security-lab/lwc/spoc

-- This is a human-readable summary of (and not a substitute for) the license.
-- You are free to:

--     Share:  copy and redistribute the material in any medium or format
--     Adapt:  remix, transform, and build upon the material

--     The licensor cannot revoke these freedoms as long as you follow
--     the license terms.

-- Under the following terms:

--     Attribution - You must give appropriate credit, provide a link to
--     the license, and indicate if changes were made. You may do so in
--     any reasonable manner, but not in any way that suggests the
--     licensor endorses you or your use.

--     NonCommercial - You may not use the material for commercial
--     purposes.

--     ShareAlike - If you remix, transform, or build upon the material,
--     you must distribute your contributions under the same license as
--     the original.

--     No additional restrictions - You may not apply legal terms or
--     technological measures that legally restrict others from doing
--     anything the license permits.
---------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.util.all;

package ctrl_pkg is

 ------------------------------------------------------------
 -- for ctrl
  subtype ds_ctrl_ty is std_logic_vector( 0 to 3 );

  function is_tag_ds 	( dom_sep : ds_ctrl_ty) 	return std_logic;  -- tag
  function is_ed_ds		( dom_sep : ds_ctrl_ty ) 	return std_logic;  -- enc/dec
  function is_ad_ds 	( dom_sep : ds_ctrl_ty ) 	return std_logic;  -- proc. ad

  subtype mode_ty is std_logic;     -- top lvl input
  function is_enc_mode	( mode : mode_ty ) 	return std_logic; -- AEAD encrypt (mode change not allowed)
  function is_dec_mode	( mode : mode_ty ) 	return std_logic; -- AEAD decrypt (mode change not allowed)

  subtype state_ty is std_logic_vector( 0 to 2 );
 	constant i_load_st    : state_ty :=  "000";
  constant l_load_st    : state_ty :=  "100";
	constant i_main_st    : state_ty :=  "010";
	constant m_main_st    : state_ty :=  "110";
	constant t_tag_st     : state_ty :=  "011";
	constant perm_st      : state_ty :=  "111";

 ------------------------------------------------------------
  -- state predicates

  function is_i_load_st 	( s : state_ty ) return std_logic;
  function is_l_load_st		( s : state_ty ) return std_logic;
  function is_i_main_st  	( s : state_ty ) return std_logic;
  function is_m_main_st 	( s : state_ty ) return std_logic;
  function is_t_tag_st		( s : state_ty ) return std_logic;
  function is_perm_st	   	( s : state_ty ) return std_logic;

  ------------------------------------------------------------
  -- counters
  ------------------------------------------------------------
  -- permutation counter
  subtype count_ty is unsigned( 7 downto 0 );
  function is_perm_one( count : count_ty ) return std_logic;
  function is_perm_done( count : count_ty ) return std_logic;
  function is_perm_r( count : count_ty ) return std_logic;


 ------------------------------------------------------------
  -- datapath operations
  ------------------------------------------------------------
  subtype  dp_op_ty is std_logic_vector( 9 downto 0 );
-- some signals used for ANDmask with onehot, rollover, counter, ... in ctrl.vhd
--  i_data muxes
  constant sel_ds_idx         : natural := 0;
  constant sel_input0_idx     : natural := 1;
  constant sel_input1_idx     : natural := 2;
-- o_data muxes
  constant sel_out_idx 		 	  : natural := 3;
  constant en_out_idx    			: natural := 4;

-- state update control
  constant sel_round_step_idx	: natural := 5;
  constant sel_load_perm_idx	: natural := 6;
  constant en_load_idx				: natural := 7;

-- constant lfsr control
  constant lfsr_c_en_idx			: natural := 8;
  constant lfsr_c_reset_idx		: natural := 9;

  -- load
  constant dp_op_1 		: dp_op_ty := (  en_load_idx   			=> '1',
  	                                   lfsr_c_reset_idx  	=> '1',
                                 		   others        			=> '0');
  -- permutation
  constant dp_op_2 		: dp_op_ty := (sel_round_step_idx   => '1',
  													         sel_load_perm_idx    => '1',
  													         lfsr_c_en_idx  		  => '1',
                                  	 others        			  => '0');
  --  proc_ad
  constant dp_op_3 		: dp_op_ty := ( sel_ds_idx          => '1',
                                      sel_input0_idx      => '1',
                                      sel_input1_idx      => '1',
  												          	sel_load_perm_idx   => '1',
  													          lfsr_c_en_idx   		=> '1',
  												          	lfsr_c_reset_idx  	=> '1',
                                  		others        			=> '0');
   -- encrypt
  constant dp_op_4 		: dp_op_ty := ( sel_ds_idx          => '1',
                                      sel_input0_idx      => '1',
                                      sel_input1_idx      => '1',
  													          en_out_idx    			=> '1',
  													          sel_load_perm_idx   => '1',
  													          lfsr_c_en_idx  		  => '1',
  												          	lfsr_c_reset_idx  	=> '1',
                        		          others        			=> '0');
  -- decrypt
  constant dp_op_5 		: dp_op_ty := ( sel_ds_idx          => '1',
                                      sel_input1_idx      => '1',
                                      en_out_idx    			=> '1',
  													          sel_load_perm_idx   => '1',
                                     lfsr_c_en_idx  	   	=> '1',
  													          lfsr_c_reset_idx  	=> '1',
                                      others        			=> '0');
  -- tag
  constant dp_op_6 		: dp_op_ty := ( sel_ds_idx          => '1',
                                      sel_load_perm_idx   => '1',
                                      lfsr_c_en_idx   		=> '1',
                                    	lfsr_c_reset_idx  	=> '1',
                                      others        			=> '0');
  constant dp_op_7 	 	: dp_op_ty := ( sel_out_idx   		  => '1',
  																		en_out_idx    			=> '1',
                        		          others 	       		  => '0');
end package;



package body ctrl_pkg is
  ------------------------------------------------------------
  -- dom sep and mode functions

 -- tag:
 -- 1000 empty AD, empty PT/CT
 -- 1010 full last block AD, empty PT/CT
 -- 1011 partial last block AD, empty PT/CT
 -- 1100 dont care AD, full last block PT/CT
 -- 1101 dont care AD, partial last block PT/CT
  function is_tag_ds		( dom_sep : ds_ctrl_ty ) 	return std_logic is
   begin
     return   dom_sep(0) ;  -- we check tag first
   end function;

  -- enc/dec: 0100 full block, 0101 partial block
  function is_ed_ds		( dom_sep : ds_ctrl_ty ) 	return std_logic is
  begin
    -- return  not dom_sep(0) and dom_sep(1) ;
    return  dom_sep(1) ;  -- because we only check this after we already know its NOT tag
  end function;

  -- proc. ad: 0010 full block, 0011 partial block
  function is_ad_ds		( dom_sep : ds_ctrl_ty ) 	return std_logic is
  begin
        -- return  not dom_sep(0) and not dom_sep(1) ;
    return   dom_sep(2) ;  -- because we only check this after we already know its NOT tag and not PT/CT
  end function;

  function is_enc_mode	( mode : mode_ty ) 	return std_logic is
  begin
    return not mode;
  end function;

  function is_dec_mode	( mode : mode_ty ) 	return std_logic is
  begin
    return mode;
  end function;

  ------------------------------------------------------------
  -- state predicates

  function is_i_load_st ( s : state_ty ) return std_logic is
  begin
    return b2x( s = i_load_st);
  end function;

  function is_l_load_st ( s : state_ty ) return std_logic is
  begin
    return b2x( s = l_load_st);
  end function;

  function is_i_main_st( s : state_ty ) return std_logic is
  begin
    return b2x( s = i_main_st);
  end function;

  function is_m_main_st( s : state_ty ) return std_logic is
  begin
    return b2x( s = m_main_st);
  end function;

  function is_t_tag_st( s : state_ty ) return std_logic is
  begin
    return b2x( s = t_tag_st);
  end function;

  function is_perm_st( s : state_ty ) return std_logic is
  begin
    return b2x( s = perm_st);
  end function;

  ------------------------------------------------------------
  function is_perm_one( count : count_ty ) return std_logic is
  begin
  return b2x(count = 1 );
  end function;

  function is_perm_r( count : count_ty ) return std_logic is
  begin
  return b2x(count = 143 );
  end function;

  function is_perm_done( count : count_ty ) return std_logic is
  begin
  return b2x(count = 144 );
  end function;



end package body;
