-- This work is licensed under a Creative Commons
-- Attribution-NonCommercial-ShareAlike 4.0 International License.
-- http://creativecommons.org/licenses/by-nc-sa/4.0

-- 2022 HW design:  Mark D. Aagaard, Nusa Zidaric
-- 2022 VHDL implementation: Nusa Zidaric
-- note overlap with ACE modules at
-- ( http://uwaterloo.ca/communications-security-lab/lwc/ace )

-- SpoC submitters/designers:
-- Riham AlTawy
-- Guang Gong
-- Morgan He
-- Ashwin Jha
-- Kalikinkar Mandal
-- Mridul Nandi
-- Raghvendra Rohit

-- http://uwaterloo.ca/communications-security-lab/lwc/spoc

-- This is a human-readable summary of (and not a substitute for) the license.
-- You are free to:

--     Share:  copy and redistribute the material in any medium or format
--     Adapt:  remix, transform, and build upon the material

--     The licensor cannot revoke these freedoms as long as you follow
--     the license terms.

-- Under the following terms:

--     Attribution - You must give appropriate credit, provide a link to
--     the license, and indicate if changes were made. You may do so in
--     any reasonable manner, but not in any way that suggests the
--     licensor endorses you or your use.

--     NonCommercial - You may not use the material for commercial
--     purposes.

--     ShareAlike - If you remix, transform, or build upon the material,
--     you must distribute your contributions under the same license as
--     the original.

--     No additional restrictions - You may not apply legal terms or
--     technological measures that legally restrict others from doing
--     anything the license permits.
---------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.spoc_128_pkg.all;
use work.ctrl_pkg.all;

entity dp is
  port
    ( clk         : in  std_logic
    ; k_count     : in  std_logic
    ; i_control 	: in 	dp_op_ty
    ; i_new_ctrl 	: in  en_ty
    ; i_pad_ctrl	: in  padxor_ty
    ; i_ds_ctrl    : in  ds_ctrl_ty
    ; i_data    : in  double_word
    ; o_data    : out double_word
   );
end entity;

architecture rtl of dp is

  signal sel_ds, sel_input0, sel_input1,
         sel_out, en_out, sel_round_step,
         sel_load_perm, en_load,
         lfsr_c_reset, lfsr_c_en          : std_logic;

  signal y_xor, y_in, z_xor, post_pad  : word_data;
  signal post_input,
         pre_round, post_round, post_xor,
         post_step_const, post_linear,
         spoc_path, spoc_perm, spoc_state : word_state_ty;
  signal dsxor                            : word;
  signal i_data_vector, o_data_vector,
         osel_data_vector                 : word_data;
  signal ctl_const, ctl_const1            : lfsr_c_output;
  signal sc0, sc1                         : std_logic_vector(lfsr_c_sz downto 0);
  signal rc0, rc1                         : std_logic;
  signal sc0c, sc1c                       : std_logic_vector(lfsr_c_sz downto 0);


begin

  u_lfsr_c :  entity work.lfsr_c port map
      (	clk, lfsr_c_en , lfsr_c_reset ,  ctl_const  );

  rc0 <= ctl_const(0);  rc1 <= ctl_const(1);
  sc0 <= ctl_const(2 to lfsr_c_sz + 2);
  sc1 <= ctl_const(lfsr_c_sz + 3 to lfsr_out_sz - 1);

  i_data_vector <= data_to_vector( i_data );
  o_data        <= vector_to_data( o_data_vector );

  sel_ds          <= i_control( sel_ds_idx );
  sel_input0      <= i_control( sel_input0_idx );
  sel_input1      <= i_control( sel_input1_idx );
  sel_out         <= i_control( sel_out_idx );
  en_out          <= i_control( en_out_idx );
  sel_round_step  <= i_control( sel_round_step_idx );
  sel_load_perm   <= i_control( sel_load_perm_idx );
  en_load         <= i_control( en_load_idx );
  lfsr_c_reset    <= i_control( lfsr_c_reset_idx );
  lfsr_c_en       <= i_control( lfsr_c_en_idx );

  ----------------------------------------------------------------------
  -- intermediate input:  do input
  z_xor(0) <= i_data_vector(0) xor spoc_state(s0_idx);
  z_xor(1) <= i_data_vector(1) xor spoc_state(s2_idx);

  y_in(0) <= i_data_vector(0) when sel_input0 = '1' else z_xor(0);
  y_in(1) <= i_data_vector(1) when sel_input0 = '1' else z_xor(1);

  y_xor(0) <= y_in(0) xor spoc_state(s1_idx);
  y_xor(1) <= y_in(1) xor spoc_state(s3_idx);

  ---------------------------------------------
  --  padding, select new or internal path
  post_pad( 0 )( 0 to 7 ) <=  y_xor( 0 )( 0 to 7 ) when i_new_ctrl(0) = '1' else
                              spoc_state ( s1_idx )( 0 to 7 );

  padbits_s1 : for i in 1 to 7 generate
    post_pad( 0 )( 8*i )  <=   y_xor( 0 )( 8*i ) when i_new_ctrl(i) = '1' else
                               spoc_state ( s1_idx )( 8*i ) xor  i_pad_ctrl(i-1);
  end generate  padbits_s1 ;

  nonpadbits_s1 : for i in 1 to 7 generate
    otherbits_s1 : for j in 1 to 7 generate
    post_pad( 0 )( 8*i + j )  <=  y_xor( 0 )( 8*i + j ) when i_new_ctrl(i) = '1'
                           else   spoc_state ( s1_idx )( 8*i + j );
    end generate otherbits_s1;
  end generate  nonpadbits_s1;


  padbits_s3 : for i in 0 to 7 generate
      post_pad( 1 )( 8*i )  <=   y_xor( 1 )( 8*i ) when i_new_ctrl(8+i) = '1'
                      else  spoc_state ( s3_idx )( 8*i ) xor  i_pad_ctrl(8+i-1);
  end generate  padbits_s3 ;

  nonpadbits_s3 : for i in 0 to 7 generate
    otherbits_s3 : for j in 1 to 7 generate
      post_pad( 1 )( 8*i + j )  <=
                                y_xor( 1 )( 8*i + j ) when i_new_ctrl(8+i) = '1'
                          else  spoc_state ( s3_idx )( 8*i + j );
    end generate otherbits_s3;
  end generate  nonpadbits_s3;

  -------------------------
  -- post input:  do input and domain separator

  post_input( s0_idx ) <=  dsxor xor spoc_state(s0_idx)    when sel_ds = '1' else
                           spoc_state(s0_idx);
  post_input( s1_idx ) <=  post_pad(0) when sel_input1 = '1' else
                           spoc_state(s1_idx);
  post_input( s2_idx ) <=  spoc_state(s2_idx);
  post_input( s3_idx ) <=  post_pad(1) when sel_input1 = '1' else
                           spoc_state(s3_idx);

  dsxor( 4 to word_sz - 1 ) <= ( others => '0' );
  dsxor( 0 to 3 )           <= i_ds_ctrl ;

  ----------------------------------------------------------------------
  --  output control
  osel_data_vector(0) <=  spoc_state ( s1_idx ) when sel_out = '1'  else --T0
                          z_xor(0) ;              --CT/PT
  osel_data_vector(1) <=  spoc_state ( s3_idx ) when sel_out = '1'  else --T1
                          z_xor(1);             --CT/PT

  o_data_vector(0) <= osel_data_vector(0) when en_out = '1' else (others => '0');
  o_data_vector(1) <= osel_data_vector(1) when en_out = '1' else (others => '0');

  ---------------------------------------------------------------------
  -- sb 64 ==> post round
  pre_round <= spoc_state;

  s1_sb_64 :  entity work.sb_64 port map
      ( pre_round( s1_idx ), rc0
      , post_round( s1_idx ) );

  s3_sb_64 :  entity work.sb_64 port map
      ( pre_round( s3_idx ), rc1
      , post_round( s3_idx ) );

  post_round( s0_idx ) <= pre_round( s0_idx );
  post_round( s2_idx ) <= pre_round( s2_idx );

  ----------------------------------------------------------------------
  -- XORs to the left ==> post xor

  post_xor( s1_idx ) <= post_round( s1_idx );
  post_xor( s3_idx ) <= post_round( s3_idx );

  post_xor( s0_idx ) <= post_round( s0_idx ) xor post_round( s1_idx );
  post_xor( s2_idx ) <= post_round( s2_idx ) xor post_round( s3_idx );

  ----------------------------------------------------------------------
  -- XOR with step constant ==> post step const

  post_step_const( s1_idx )           <= post_xor( s1_idx );
  post_step_const( s3_idx )           <= post_xor( s3_idx );

  post_step_const( s0_idx)( 0 to 55)  <= not post_xor( s0_idx )( 0 to 55  );
  post_step_const( s0_idx)( 56 to 63) <= post_xor( s0_idx )( 56 to 63 ) xor sc0;

  post_step_const( s2_idx)( 0 to 55)  <= not post_xor( s2_idx )( 0 to 55  );
  post_step_const( s2_idx)( 56 to 63) <= post_xor( s2_idx )( 56 to 63 ) xor sc1;

  ------------------------------------------------------------
  -- post liear layer pi = (1,2,3,0) ==> post linear

  post_linear( s0_idx ) <= post_step_const( s1_idx );
  post_linear( s1_idx ) <= post_step_const( s2_idx );
  post_linear( s2_idx ) <= post_step_const( s3_idx );
  post_linear( s3_idx ) <= post_step_const( s0_idx );

  ------------------------------------------------------------
  -- update state

  spoc_perm  <= post_linear  when sel_round_step = '1'  else post_round ;
  spoc_path  <= post_input   when sel_ds = '1'          else spoc_perm;

  process
  begin
    wait until rising_edge(clk);
      if sel_load_perm = '1' then
        spoc_state <= spoc_path; -- every clk cycle
      else
        if en_load = '1'   then     -- load
          if k_count = '1' then
            spoc_state( s1_idx ) <= i_data_vector(0);
            spoc_state( s3_idx ) <= i_data_vector(1);
          else
            spoc_state( s0_idx ) <= i_data_vector(0);
            spoc_state( s2_idx ) <= i_data_vector(1);
          end if;
        end if;
      end if;
  end process;

end architecture;
