-- This work is licensed under a Creative Commons
-- Attribution-NonCommercial-ShareAlike 4.0 International License.
-- http://creativecommons.org/licenses/by-nc-sa/4.0

-- SpoC-128_sLiSCP-light-[256]

-- 2022 HW design:  Mark D. Aagaard, Nusa Zidaric
-- 2022 VHDL implementation: Nusa Zidaric
-- note overlap with ACE modules at
-- ( http://uwaterloo.ca/communications-security-lab/lwc/ace )

-- SpoC submitters/designers:
-- Riham AlTawy
-- Guang Gong
-- Morgan He
-- Ashwin Jha
-- Kalikinkar Mandal
-- Mridul Nandi
-- Raghvendra Rohit

-- http://uwaterloo.ca/communications-security-lab/lwc/spoc

-- This is a human-readable summary of (and not a substitute for) the license.
-- You are free to:

--     Share:  copy and redistribute the material in any medium or format
--     Adapt:  remix, transform, and build upon the material

--     The licensor cannot revoke these freedoms as long as you follow
--     the license terms.

-- Under the following terms:

--     Attribution - You must give appropriate credit, provide a link to
--     the license, and indicate if changes were made. You may do so in
--     any reasonable manner, but not in any way that suggests the
--     licensor endorses you or your use.

--     NonCommercial - You may not use the material for commercial
--     purposes.

--     ShareAlike - If you remix, transform, or build upon the material,
--     you must distribute your contributions under the same license as
--     the original.

--     No additional restrictions - You may not apply legal terms or
--     technological measures that legally restrict others from doing
--     anything the license permits.
---------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

package spoc_128_pkg is

  -- for the lfsr_c to produce the constants
  constant lfsr_c_sz    :  integer := 7;
  constant lfsr_out_sz  :  integer := 18;
  subtype lfsr_c_output is std_logic_vector(0 to lfsr_out_sz - 1);

  ------------------------------------------------------------

  constant half_word_sz   : natural := 32;
  constant word_sz        : natural := 2*half_word_sz;
  constant double_word_sz : natural := 2*word_sz;

  subtype half_word     is std_logic_vector( 0 to half_word_sz - 1 );
  subtype word          is std_logic_vector( 0 to word_sz - 1 );
  subtype double_word   is std_logic_vector( 0 to double_word_sz - 1 );
  type word_vector  	  is array( natural range <> ) of word;
  type half_word_vector is array( natural range <> ) of half_word;

  ------------------------------------------------------------
  constant state_sz          : natural := 256;
  constant word_max_idx      : natural := state_sz / word_sz - 1;
  constant half_word_max_idx : natural := state_sz / half_word_sz - 1;

  constant key_sz            : natural := 128;
  constant nonce_sz          : natural := 128;

  subtype word_state_ty is word_vector ( 0 to word_max_idx );
  constant s0_idx : natural := 0;
  constant s1_idx : natural := 1;
  constant s2_idx : natural := 2;
  constant s3_idx : natural := 3;


  subtype half_word_data is  half_word_vector ( 0 to 1 );
  subtype word_data is  word_vector ( 0 to 1 );

  function vector_to_data ( st : word_data ) return double_word;
  function data_to_vector ( st : double_word ) return word_data;
  ----------------------------------------------------------------------
  -- for padding
  subtype pad_ty is std_logic_vector(3 downto 0 );
  subtype en_ty is std_logic_vector( 0 to 15 );   		-- en  ctrl for decryption
  subtype padxor_ty is std_logic_vector( 0 to 14 );    -- pad ctrl for dectryption - pad detection

end package;


package body spoc_128_pkg is

  ------------------------------------------------------------
  -- state functions
  ------------------------------------------------------------
  function data_to_vector( st : double_word )
    return word_data
  is
    variable z : word_data;
  begin
    z(0) := st(0 to word_sz - 1);
    z(1) := st(word_sz to double_word_sz - 1);
    return z;
  end function;


  function vector_to_data( st : word_data )
    return double_word
  is
    variable z : double_word;
  begin
    z(0 to word_sz - 1) := st(0);
    z(word_sz to double_word_sz - 1) := st(1);
    return z;
  end function;


end package body;
