-- This work is licensed under a Creative Commons
-- Attribution-NonCommercial-ShareAlike 4.0 International License.
-- http://creativecommons.org/licenses/by-nc-sa/4.0

-- 2022 HW design:  Mark D. Aagaard, Nusa Zidaric
-- 2022 VHDL implementation: Nusa Zidaric
-- note overlap with ACE modules at
-- ( http://uwaterloo.ca/communications-security-lab/lwc/ace )

-- SpoC submitters/designers:
-- Riham AlTawy
-- Guang Gong
-- Morgan He
-- Ashwin Jha
-- Kalikinkar Mandal
-- Mridul Nandi
-- Raghvendra Rohit

-- http://uwaterloo.ca/communications-security-lab/lwc/spoc

-- This is a human-readable summary of (and not a substitute for) the license.
-- You are free to:

--     Share:  copy and redistribute the material in any medium or format
--     Adapt:  remix, transform, and build upon the material

--     The licensor cannot revoke these freedoms as long as you follow
--     the license terms.

-- Under the following terms:

--     Attribution - You must give appropriate credit, provide a link to
--     the license, and indicate if changes were made. You may do so in
--     any reasonable manner, but not in any way that suggests the
--     licensor endorses you or your use.

--     NonCommercial - You may not use the material for commercial
--     purposes.

--     ShareAlike - If you remix, transform, or build upon the material,
--     you must distribute your contributions under the same license as
--     the original.

--     No additional restrictions - You may not apply legal terms or
--     technological measures that legally restrict others from doing
--     anything the license permits.
---------------------------

LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.numeric_std.all;
USE STD.textio.all;
USE ieee.std_logic_textio.all;

use work.util_unsynth.all;
use work.util.all;
use work.tb_spoc_unsynth.all;
use work.spoc_128_pkg.all;
use work.ctrl_pkg.all;
use work.tb_pkg.all;

ENTITY tb_spoc IS
  generic ( period : real    := 10.0
          ; ED     : std_logic_vector(0 to 1) := "11" -- Encryption | Decryption
          ; S      : std_logic_vector(0 to 1) := "01" -- stall flag: 00 - none, 01 - stall enc, 10 - stall dec, 11 stall other
         );
END tb_spoc;

ARCHITECTURE main OF tb_spoc IS

  file simo_file      : text open write_mode is simo_file_path;
  file stim_file      : text;

  constant clk_period   : time  := period * 1 ns;
  constant hold         : time  := clk_period / 5;
  constant after_reset  : time  := 0 * clk_period;  -- delay between reset and loading
  constant load_delay   : time  := 0 * clk_period;  -- delay between K, N data
  constant procad_delay : time  := 0 * clk_period;  -- AD data delay
  constant stall_short_delay  : time  := 3   * clk_period;  -- stall short delay
  constant stall_long_delay   : time  := 300 * clk_period;  -- stall long delay
  constant stall_period       : time  := 5   * clk_period;
  constant enc_delay    : time  := 10 * clk_period;  -- encryption data delay
  constant dec_delay    : time  := 10 * clk_period;  -- decryption data delay
  constant mode_delay   : time  := 50 * clk_period;  -- delay between two modes



  signal clk, reset : std_logic := '0';
  signal stall : boolean := false;

  signal tag_data : double_word;
  signal tag_valid   : std_logic := '0';

  type tb_state_ty is (tbLoad,  tbProcAD, tbEncrypt, tbDecrypt,   tbTag,   tbNull);
  signal tb_state : tb_state_ty;

  signal pre_data, post_data    : double_word;
  signal pre_valid, pre_ready   : std_logic;
  signal post_valid, post_ready : std_logic;
  signal pre_ds_ctrl, final_ds_ctrl   : ds_ctrl_ty;
  signal len, len_tracker : std_logic_vector(0 to 3);

  signal pre_mode        : mode_ty;
  signal pre_padding     : pad_ty;
  signal bits_unpadded   : integer;
  signal mode_stim				 : stim_ty;

BEGIN



  post_reading_proc : process -- reading the tb/sim output , writing the simo file
    variable msg , empty   : line;
    variable done, skip_write, skip_empty  : boolean;
    variable post_data_var : double_word;

  begin

    skip_write := false;
    done := false;
    while not(done) loop
      wait until rising_edge(clk);
      done := (post_valid = '1' or tag_valid = '1' );
    end loop;

    -- post_data_var := post_data;
	    if tb_state = tbEncrypt then
			 if 	bits_unpadded >= 0  then
            skip_empty := true;
  			    write(  msg, "C " & to_hex_string_normal( post_data( 0 to bits_unpadded-1 ) ));
			 else
	       		skip_write := true;
            skip_empty := true;
	      end if;
	    elsif tb_state = tbDecrypt then
		    if 	bits_unpadded >= 0 then
            skip_empty := true;
	     			write(  msg, "P " & to_hex_string_normal( post_data( 0 to bits_unpadded-1 ) ));
			 else
	       		skip_write := true;
            skip_empty := true;
	     end if;

	    elsif tb_state = tbTag then
        if tag_valid = '1' then
          write(  msg, "T " & to_hex_string_normal( tag_data ));
          if final_ds_ctrl(1) = '0' then -- empty PT/CT
            if   mode_stim = PLAINTEXT_STIM  then
              write(empty, 'C');  write(empty, ' ');
              skip_empty := false;
            elsif mode_stim = CIPHERTEXT_STIM then
              write(empty, 'P');   write(empty, ' ');
              skip_empty := false;
           end if;
          else
            skip_empty := true;
          end if;
	      else
	        skip_write := true;
          skip_empty := true;
	      end if;
	    else
	      skip_write := true;
        skip_empty := true;
	    end if;

    if not(skip_empty) then writeline( simo_file, empty); end if;
    if not(skip_write) then writeline( simo_file, msg); end if;

  end process;

  post_stalling_proc : process

  begin

      if S = "01" then
        if   tb_state = tbEncrypt then
        stall <= true;
        wait for stall_short_delay;
      end if;
      stall <= false;
      wait for stall_period;
    elsif S = "10" then
        if   tb_state = tbDecrypt then
        stall <= true;
        wait for stall_short_delay;
      end if;
      stall <= false;
      wait for stall_period;
    elsif S = "11" then
        if  tb_state = tbProcAD or    tb_state = tbTag    then
        stall <= true;
        --wait for stall_long_delay;
        wait for stall_short_delay;
      end if;
      stall <= false;
      wait for stall_period;

    else
       stall <= false;
       wait for stall_period;
      end if;

  end process;

  post_ready <= '0' when stall else '1';
  clk <= not(clk) after clk_period / 2;

   -- Instantiate the Unit Under Test (UUT)
  uut: entity work.spoc_128 port map
    (  clk       => clk
    ,  reset     => reset
    ,  pre_mode    => pre_mode
    ,  pre_ds_ctrl => pre_ds_ctrl
    ,  pre_padding => pre_padding
    ,  pre_valid   => pre_valid
    ,  post_ready  => post_ready
    ,  pre_data    => pre_data
    ,  post_data    => post_data
    ,  post_valid   => post_valid
    ,  pre_ready    => pre_ready
     );

  stim_proc: process
    variable ds_ctrl  : ds_ctrl_ty;
    variable target_stim					: stim_ty;
    variable key_rd, nonce_rd, data 			: double_word;

  begin

    if ED(0) = '1' then
      ------------------------------------------------------------------------------
      ---------------------------------- Encryption --------------------------------
      ------------------------------------------------------------------------------

      tag_data <= (others => 'X');

      pre_valid  <= '0';
      pre_padding  <= (others => '0');
      pre_ds_ctrl  <= (others => '0');
      final_ds_ctrl  <= (others => '0');
      ds_ctrl  := (others => '0');

      pre_mode <= '0';
      tb_state <= tbNull;
      target_stim := NULL_STIM;
      mode_stim <= PLAINTEXT_STIM;
      -------------------------------- Reset -------------------------------


      drive_reset( clk, reset, hold );
      wait until rising_edge(clk);

      -------------------------------- Load --------------------------------
      report( "LOADING PHASE" );
      tb_state <= tbLoad;

      read_key_nonce( key_rd, nonce_rd, stim_file_path );


      drive_data( clk, key_rd, ds_ctrl, hold, pre_ready, pre_data, pre_valid,
                  pre_ds_ctrl );    wait for load_delay; --wait until rising_edge(clk);
      drive_data( clk, nonce_rd, ds_ctrl, hold, pre_ready, pre_data, pre_valid,
                  pre_ds_ctrl );    wait for load_delay; --wait until rising_edge(clk);

      ---------------------------------- ProcAD --------------------------------
      report( "AD PROCESSING PHASE" );
      tb_state <= tbProcAD;

		  target_stim := AD_STIM;
      drive_all(target_stim, stim_file_path, hold, procad_delay, clk, pre_ready,
                post_valid, pre_ds_ctrl, pre_data, pre_valid, pre_padding,
                bits_unpadded, pre_ds_ctrl);

      ---------------------------------- Encrypt ------------------------------
      report( "ENCRYPTION PHASE" );
      tb_state <= tbEncrypt;

		  target_stim := PLAINTEXT_STIM;
      drive_all(target_stim, stim_file_path, hold, procad_delay, clk, pre_ready,
                post_valid, pre_ds_ctrl, pre_data, pre_valid, pre_padding,
                bits_unpadded,  pre_ds_ctrl);
      final_ds_ctrl <= pre_ds_ctrl;


     ---------------------------------- Final --------------------------------
      report( "TAG PERMUTATION" );
      tb_state <= tbTag;
      data := (others => '0');
      ds_ctrl  := "1000";
      drive_data( clk, data, ds_ctrl, hold, pre_ready, pre_data, pre_valid,
                  pre_ds_ctrl );

      ---------------------------------- Tag ----------------------------------
      report( "TAG PHASE   "  );
    	wait until  post_ready ='1' and post_valid = '1' and rising_edge(clk);
      tag_valid <=   '1';
      tag_data <= post_data;
    	report( "tag vector was read for  " & to_hex_string_normal( post_data ) );

      wait until rising_edge(clk); wait for hold;
      report("tag is: " & to_hex_string_normal( tag_data ) );
      tag_valid <= '0';
      wait for mode_delay;

    end if;


    if ED(1) = '1' then
      ------------------------------------------------------------------------------
      ---------------------------------- Decryption --------------------------------
      ------------------------------------------------------------------------------
      tag_data <= (others => 'X');

      pre_valid  <= '0';
      pre_padding  <= (others => '0');
      pre_ds_ctrl  <= (others => '0');
      final_ds_ctrl  <= (others => '0');
      ds_ctrl  := (others => '0');

      pre_mode <= '1';
      tb_state <= tbNull;
      target_stim := NULL_STIM;
      mode_stim <= CIPHERTEXT_STIM;

      -------------------------------- Load --------------------------------
      report( "LOADING PHASE" );
      tb_state <= tbLoad;

      read_key_nonce( key_rd, nonce_rd, stim_file_path );

      ds_ctrl := "0000";
      drive_data( clk, key_rd, ds_ctrl, hold, pre_ready, pre_data, pre_valid,
                  pre_ds_ctrl );        wait for load_delay;
      drive_data( clk, nonce_rd, ds_ctrl, hold, pre_ready, pre_data, pre_valid,
                  pre_ds_ctrl );        wait for load_delay;

      ---------------------------------- ProcAD --------------------------------
      report( "AD PROCESSING PHASE" );
      tb_state <= tbProcAD;
		  target_stim := AD_STIM;
      drive_all(target_stim, stim_file_path, hold, procad_delay, clk, pre_ready,
                post_valid, pre_ds_ctrl, pre_data, pre_valid, pre_padding,
                bits_unpadded,  pre_ds_ctrl);

      ---------------------------------- Decrypt ------------------------------
      report( "DECRYPTION PHASE" );
      tb_state <= tbDecrypt;

		  target_stim := CIPHERTEXT_STIM;
      drive_all(target_stim, stim_file_path, hold, procad_delay, clk, pre_ready,
                post_valid, pre_ds_ctrl, pre_data, pre_valid, pre_padding,
                bits_unpadded,  pre_ds_ctrl);
      final_ds_ctrl <= pre_ds_ctrl;

      ---------------------------------- Final --------------------------------
      report( "TAG PERMUTATION" );
      tb_state <= tbTag;
      data := (others => '0');
      ds_ctrl  := "1000";
      drive_data( clk, data, ds_ctrl, hold, pre_ready, pre_data, pre_valid,
                  pre_ds_ctrl );

      ---------------------------------- Tag ----------------------------------
      report("TAG PHASE   "  );
    	wait until  post_ready ='1' and post_valid = '1' and rising_edge(clk);
      tag_valid <=   '1';
      tag_data <= post_data;
    	report("tag vector was read for  " & to_hex_string_normal( post_data ) );
      wait until rising_edge(clk); wait for hold;
      report("tag is: " & to_hex_string_normal( tag_data ) );
      tag_valid <= '0';
      wait for mode_delay;

    end if;

    assert false
    report ("SIMULATION IS FINISHED")
    severity failure;

  end process;
end;
