-- This work is licensed under a Creative Commons
-- Attribution-NonCommercial-ShareAlike 4.0 International License.
-- http://creativecommons.org/licenses/by-nc-sa/4.0

-- 2022 HW design:  Mark D. Aagaard, Nusa Zidaric
-- 2022 VHDL implementation: Nusa Zidaric
-- note overlap with ACE modules at
-- ( http://uwaterloo.ca/communications-security-lab/lwc/ace )

-- SpoC submitters/designers:
-- Riham AlTawy
-- Guang Gong
-- Morgan He
-- Ashwin Jha
-- Kalikinkar Mandal
-- Mridul Nandi
-- Raghvendra Rohit

-- http://uwaterloo.ca/communications-security-lab/lwc/spoc

-- This is a human-readable summary of (and not a substitute for) the license.
-- You are free to:

--     Share:  copy and redistribute the material in any medium or format
--     Adapt:  remix, transform, and build upon the material

--     The licensor cannot revoke these freedoms as long as you follow
--     the license terms.

-- Under the following terms:

--     Attribution - You must give appropriate credit, provide a link to
--     the license, and indicate if changes were made. You may do so in
--     any reasonable manner, but not in any way that suggests the
--     licensor endorses you or your use.

--     NonCommercial - You may not use the material for commercial
--     purposes.

--     ShareAlike - If you remix, transform, or build upon the material,
--     you must distribute your contributions under the same license as
--     the original.

--     No additional restrictions - You may not apply legal terms or
--     technological measures that legally restrict others from doing
--     anything the license permits.
---------------------------

use std.textio.all;
library ieee;
use ieee.math_real.all;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_textio.all;

use work.util_unsynth.all;
use work.spoc_128_pkg.all;
use work.ctrl_pkg.all;

package tb_spoc_unsynth is

  constant max_data_sz     : integer := 256;
  -- ******************************************** --
  type stim_ty is ( KEY_STIM, NONCE_STIM, INPUT_STIM, OUTPUT_STIM, AD_STIM,
  PLAINTEXT_STIM, CIPHERTEXT_STIM, EOF_STIM, NULL_STIM );

  function to_hex_string_normal( v : std_logic_vector ) return string;
  -- ******************************************** --
  function is_stim( c : character ) return boolean;
  function chr_to_stim( chr : character ) return stim_ty;
  -- ******************************************** --
  procedure read_next_stim_data_line_normal
    ( file f   : text;
      stim     : out stim_ty;
      data     : out std_logic_vector;
      num_bits : out natural );

  procedure hread_dir_normal
    ( ln       : inout line;
      result   : out   std_logic_vector;
      num_bits : out   natural
    );

  function pad ( v : std_logic_vector ) return std_logic_vector;

  procedure drive_reset
    ( signal   clk   : in std_logic
    ; signal   reset : out std_logic
    ; constant hold  : in time
    );

  procedure read_key_nonce
    ( key_rd       : out std_logic_vector
    ; nonce_rd     : out std_logic_vector
    ; stim_file_path : in string
    );

  procedure drive_data
    ( signal   clk     : in std_logic
    ; variable data    : in std_logic_vector
    ; variable ds_ctrl : in ds_ctrl_ty
    ; constant hold    : in time
    ; signal   pre_ready  : in std_logic
    ; signal   pre_data   : out std_logic_vector
    ; signal   pre_valid  : out std_logic
    ; signal pre_ds_ctrl  : out ds_ctrl_ty  -- current block ctrl
    );

    procedure drive_pt_ct
      ( signal   clk        : in std_logic
      ; variable data       : in std_logic_vector
      ; variable ds_ctrl    : in ds_ctrl_ty
      ; constant hold       : in time
      ; signal   pre_ready  : in std_logic
      ; signal   post_valid : in std_logic
      ; signal   pre_data   : out std_logic_vector
      ; signal   pre_valid  : out std_logic
      ; signal pre_ds_ctrl  : out ds_ctrl_ty  -- current block ctrl
      );

  procedure drive_all
    ( constant target_stim: in stim_ty
    ; stim_file_path      : in string
    ; constant hold       : in time
    ; constant delay      : in time
    ; signal clk          : in std_logic
    ; signal pre_ready    : in std_logic
    ; signal post_valid   : in std_logic
    ; signal old_ds_ctrl  : in ds_ctrl_ty   -- needed tag_ctrl
    ; signal pre_data     : out std_logic_vector
    ; signal pre_valid    : out std_logic
    ; signal pre_padding  : out pad_ty
    ; signal bits_unpadded: out integer
    ; signal pre_ds_ctrl  : out ds_ctrl_ty  -- current block ctrl
    );

end package;

package body tb_spoc_unsynth is

  procedure drive_reset
    ( signal   clk   : in std_logic
    ; signal   reset : out std_logic
    ; constant hold  : in time
    )
  is
  begin

    wait until rising_edge(clk);
    wait for hold;
    reset <= '1';
    wait until rising_edge(clk);
    reset <= '0';

  end procedure;

  procedure read_key_nonce
    ( key_rd          : out std_logic_vector
    ; nonce_rd        : out std_logic_vector
    ; stim_file_path    : in string
    )   is
    variable stim_rd   : stim_ty;
    file     stim_file : text;
    variable num_bits  : natural;
  begin

    file_close(stim_file);
    file_open(stim_file, stim_file_path, read_mode);

    stim_rd := NULL_STIM;
    while stim_rd /= KEY_STIM and stim_rd /= NONCE_STIM and stim_rd /= EOF_STIM
      loop
        read_next_stim_data_line_normal( stim_file, stim_rd, key_rd, num_bits );
    end loop;

    if stim_rd = KEY_STIM then
      if num_bits /= 128 then
        file_close(stim_file);
        assert false
        report( "Key length = " & natural'image(num_bits) &
                " bits, must be 128 bits")
        severity failure;
      else
        report( "Key length = " & natural'image(num_bits) & " bits");
      end if;
    else
      file_close(stim_file);
      assert false
      report ("Key isn't found in file " & stim_file_path)
      severity failure;
    end if;

    file_close(stim_file);
    file_open(stim_file, stim_file_path, read_mode);

    stim_rd := NULL_STIM;
    while stim_rd /= NONCE_STIM and stim_rd /= EOF_STIM loop
      read_next_stim_data_line_normal( stim_file, stim_rd, nonce_rd, num_bits );
    end loop;

    if stim_rd = NONCE_STIM then
      if num_bits /= 128 then
        file_close(stim_file);
        assert false
        report( "Nonce length = " & natural'image(num_bits) &
                " bits, must be 128 bits")
        severity failure;
      else
        report( "Nonce length = " & natural'image(num_bits) & " bits");
      end if;
    else
      file_close(stim_file);
      assert false
      report ("Nonce isn't found in file " & stim_file_path)
      severity failure;
    end if;
    file_close(stim_file);
  end procedure;

  procedure drive_data
    ( signal   clk      : in std_logic
    ; variable data     : in std_logic_vector
    ; variable ds_ctrl  : in ds_ctrl_ty
    ; constant hold     : in time
    ; signal   pre_ready  : in std_logic
    ; signal   pre_data   : out std_logic_vector
    ; signal   pre_valid  : out std_logic
    ; signal pre_ds_ctrl  : out ds_ctrl_ty  -- current block ctrl
    )   is
  begin
    wait for hold;
    pre_data     <= data;
    pre_ds_ctrl  <= ds_ctrl;
    pre_valid    <= '1';
    wait until pre_ready = '1' and rising_edge(clk);
--    wait for hold;  --   causes transition to pFinal instead pInit
    pre_valid  <= '0';

  end procedure;

  procedure drive_pt_ct
    ( signal   clk      : in std_logic
    ; variable data     : in std_logic_vector
    ; variable ds_ctrl  : in ds_ctrl_ty
    ; constant hold     : in time
    ; signal   pre_ready  : in std_logic
    ; signal   post_valid : in std_logic
    ; signal   pre_data   : out std_logic_vector
    ; signal   pre_valid  : out std_logic
    ; signal pre_ds_ctrl  : out ds_ctrl_ty  -- current block ctrl
    )  is
  begin
    wait for hold;
    pre_data     <= data;
    pre_ds_ctrl  <= ds_ctrl;
    pre_valid    <= '1';
    wait until post_valid = '1' and rising_edge(clk);
    pre_valid  <= '0';
  end procedure;

  procedure drive_all
    ( constant target_stim : in stim_ty
    ; stim_file_path       : in string
    ; constant hold        : in time
    ; constant delay       : in time
    ; signal clk           : in std_logic
    ; signal pre_ready     : in std_logic
    ; signal post_valid    : in std_logic
    ; signal old_ds_ctrl   : in ds_ctrl_ty   -- needed tag_ctrl
    ; signal pre_data      : out std_logic_vector
    ; signal pre_valid     : out std_logic
    ; signal pre_padding   : out pad_ty
    ; signal bits_unpadded : out integer  -- needed for writing the stim output ( cut Ctxt/Ptxt to correct length)
    ; signal pre_ds_ctrl   : out ds_ctrl_ty  -- current block ctrl
    )  is
    variable stim_rd  : stim_ty; --returned by the read_next_stim_data_line_normal
    variable num_bits  : natural;
    variable data, data128: double_word;
    variable ds_ctrl  : ds_ctrl_ty;
    file     stim_file : text;
  begin
    pre_padding <= (others => '0');
    bits_unpadded  <= 128;
    file_close(stim_file);
    file_open(stim_file, stim_file_path, read_mode);

    stim_rd := NULL_STIM;
    while stim_rd /= target_stim and stim_rd /= EOF_STIM loop
      read_next_stim_data_line_normal( stim_file, stim_rd, data128, num_bits );
    end loop;

    while stim_rd /= EOF_STIM loop
      report( stim_ty'image(stim_rd)  & " - data length = " &
              natural'image(num_bits) & " bits");
		if num_bits = 0 then -- empty block - dont drive anything !
        ds_ctrl := "0000"; --empty block
        report( "empty block" );
    elsif num_bits < 128 then 			-- partial block
-- permutation
        data := (others => '0');
        ds_ctrl  := "0000" ;
        drive_data( clk, data, ds_ctrl, hold, pre_ready, pre_data, pre_valid,
                    pre_ds_ctrl );
        report( "PERM" );
-- followed by main
   	    if  stim_rd = PLAINTEXT_STIM	 or   stim_rd = CIPHERTEXT_STIM then
          ds_ctrl := "0101"; --partial PT/CT
          data := data128;
          data := pad( data );
          bits_unpadded  <= num_bits  ;
          pre_padding(3 downto 0) <= std_logic_vector(to_unsigned( num_bits/8,4));
          drive_pt_ct( clk, data, ds_ctrl, hold, pre_ready, post_valid, pre_data,
                       pre_valid, pre_ds_ctrl );
          report( "Partial PT/CT" );
        elsif stim_rd = AD_STIM then
          ds_ctrl := "0011"; --partial AD
          data := data128;
          data := pad( data );
          bits_unpadded  <= num_bits  ;
          pre_padding(3 downto 0) <= std_logic_vector(to_unsigned( num_bits/8 , 4) );
          drive_data( clk, data, ds_ctrl, hold, pre_ready, pre_data, pre_valid,
                      pre_ds_ctrl );
          report( "Partial AD" );
        end if;
    elsif num_bits = 128 then 		-- full block
  -- permutation
          data := (others => '0');
          ds_ctrl  := "0000" ;
          drive_data( clk, data, ds_ctrl, hold, pre_ready, pre_data, pre_valid,
                      pre_ds_ctrl );
          report( "PERM" );
-- followed by main
        if  stim_rd = PLAINTEXT_STIM	 or   stim_rd = CIPHERTEXT_STIM then
          ds_ctrl := "0100"; --full PT/CT
          data :=  data128;
           bits_unpadded  <= num_bits;
           pre_padding <= (others => '0');
           drive_pt_ct(clk, data, ds_ctrl, hold, pre_ready, post_valid, pre_data,
                       pre_valid, pre_ds_ctrl );
          report( "Full PT/CT" );
        elsif stim_rd = AD_STIM then
          ds_ctrl := "0010"; --full AD
          data :=  data128;
          bits_unpadded  <= num_bits;
          pre_padding <= (others => '0');
          drive_data( clk, data, ds_ctrl, hold, pre_ready, pre_data, pre_valid,
                      pre_ds_ctrl );
          report( "Full AD" );
        end if;
      end if;
--- when stim data length > 128 will read it in the next line:
      stim_rd := NULL_STIM;
      while stim_rd /= target_stim and stim_rd /= EOF_STIM loop
        read_next_stim_data_line_normal( stim_file, stim_rd, data128, num_bits );
      end loop;
    end loop;
    file_close(stim_file);
  end procedure;

  -- ******************************************** --
  -- ******************************************** --
  -- pad with 10* rule
  function pad ( v : std_logic_vector ) return std_logic_vector is
    variable z : std_logic_vector(v'low to v'high);
  begin
    report( "PADDING for v = " & to_string(v) );
    for i in v'high downto v'low + 1 loop
      if v(i) /= '1' and v(i) /= '0' then
			if v(i-1) = '1' or v(i-1) = '0' then
	 		 z(i) := '1';
			else
	  		 z(i) := '0';
		end if;
      else
			z(i) := v(i);
      end if;
    end loop;
    z(0) := v(0);
    return(z);
  end function;
  -- ******************************************** --
  -- ******************************************** --
  function to_hex_string_normal( v : std_logic_vector ) return string is
  begin
    if v'length = 0 then
      return "";
    elsif v'length <= 4 then
      return to_hex_char( rev( v ) ) & "";
    else
      if v'ascending then

        return to_hex_char( rev( v( v'low to v'low+3)) )
               & to_hex_string_normal( v(v'low+4 to v'high) );
      else

        return to_hex_char( rev( v( v'high downto v'high-3)) )
               & to_hex_string_normal( v(v'high-4 downto v'low) );
      end if;
    end if;
  end function;
  -- ******************************************** --
  -- ******************************************** --
  function is_stim( c : character ) return boolean is
  begin
    return c = 'K' or c = 'N' or c = 'I' or c = 'O' or c = 'A' or c = 'P' or c = 'C' or c='H';
  end function;
  -- ******************************************** --
  -- ******************************************** --
  function chr_to_stim( chr : character ) return stim_ty is
  begin
    case chr is
      when 'K' => return KEY_STIM;
      when 'N' => return NONCE_STIM;
      when 'I' => return INPUT_STIM;
      when 'O' => return OUTPUT_STIM;
      when 'A' => return AD_STIM;
      when 'P' => return PLAINTEXT_STIM;
      when 'C' => return CIPHERTEXT_STIM;

      when others =>
        assert false
          report ("chr_to_stim: not a stim: "& character'image( chr ))
          severity error;
        return EOF_STIM;
    end case;
  end function;
  -- ******************************************** --
  -- ******************************************** --
   procedure read_next_stim_data_line_normal
    ( file f   : text;
      stim     : out stim_ty;
      data     : out std_logic_vector;
      num_bits : out natural )
  is
    variable found
           , ok            : boolean;
    variable input_line    : line;
    variable chr           : character;
    variable tmp_stim      : stim_ty;
    variable tmp_data      : std_logic_vector( data'range );
    variable tmp_num_bits  : natural;
  begin
    -- read lines until hit one where first non-blank character is a stim charcter
    found := false;
    while not found and not endfile( f ) loop
      readline( f, input_line );
      read( input_line, chr, good => ok );
      -- skip blanks and comments
      found := ok and is_stim(chr);
    end loop;
    if not found then
      tmp_stim  := EOF_STIM;
      tmp_data := (others => 'U');
    else
      tmp_stim  := chr_to_stim( chr );
      hread_dir_normal( input_line, tmp_data, tmp_num_bits );

    end if;
    if tmp_num_bits > 0 then
      report( "read_next_stim_data: --> "& stim_ty'image(tmp_stim) &" "&
              to_string( firstn( tmp_data, tmp_num_bits ) ) );
    else
      report( "read_next_stim_data: --> "& stim_ty'image(tmp_stim) &" 0 bits");
    end if;
    stim     := tmp_stim;
    data     := tmp_data;
    num_bits := tmp_num_bits;
  end procedure;
  -- ******************************************** --
  -- ******************************************** --
 procedure hread_dir_normal
    ( ln       : inout line;
      result   : out   std_logic_vector;
      num_bits : out   natural
    )
  is
    variable ok           : boolean;
    variable chr          : character;
    variable str          : string(1 to result'length/4);
    variable tmp_result   : std_logic_vector(0 to result'length-1);
    variable num_chars,
             tmp_num_bits : natural := 0;
  begin

    if result'length mod 4 /= 0 then
      assert false
        report
          "hread_dir error: trying to read vector " &
             "with non multiple of 4 length";
      return;
    end if;

    loop                                    -- skip white space
      read(ln, chr, ok);
      exit when not ok or ((chr /= ' ') and (chr /= CR) and (chr /= HT));
    end loop;
    if not ok then
      assert false
        report ("hread_dir: end of line")         severity warning;
      return;
    else
     -- report ("hread_dir: char0="& chr );
    end if;

    num_chars        := 1;
    str( num_chars ) := chr;

    while ok loop
      read(ln, chr, ok);
      if ok then
        num_chars        := num_chars + 1;
        str( num_chars ) := chr;
      end if;
    end loop;

    tmp_num_bits := num_chars * 4;
    tmp_result   := (others => 'U');
    if result'ascending then
      tmp_result( 0 to tmp_num_bits - 1) :=
        hex_string_to_dn_vector( str( 1 to num_chars ) );   -- <––– changed from up to dn (change 1/2)
                                                            -- HEX(0) should be treated MSB, HEX(3) should be treated as LSB
                                                            -- for correct interpretation of HEX
    else
      tmp_result( 0 to tmp_num_bits - 1) :=
        hex_string_to_up_vector( str( 1 to num_chars ) );   -- <––– changed from up to dn (change 2/2)
                                                            -- HEX(0) should be treated MSB, HEX(3) should be treated as LSB
                                                            -- for correct interpretation of HEX
    end if;

    -- report( "hread_dir:  " & str &" -->"&
    --        to_string( tmp_result( 0 to tmp_num_bits - 1 ) ) );

    result   := tmp_result;
    num_bits := tmp_num_bits;
  end procedure;

end package body;
