-- This work is licensed under a Creative Commons
-- Attribution-NonCommercial-ShareAlike 4.0 International License.
-- http://creativecommons.org/licenses/by-nc-sa/4.0

-- 2022 HW design:  Mark D. Aagaard, Nusa Zidaric
-- 2022 VHDL implementation: Nusa Zidaric
-- note overlap with ACE modules at
-- ( http://uwaterloo.ca/communications-security-lab/lwc/ace )

-- SpoC submitters/designers:
-- Riham AlTawy
-- Guang Gong
-- Morgan He
-- Ashwin Jha
-- Kalikinkar Mandal
-- Mridul Nandi
-- Raghvendra Rohit

-- http://uwaterloo.ca/communications-security-lab/lwc/spoc

-- This is a human-readable summary of (and not a substitute for) the license.
-- You are free to:

--     Share:  copy and redistribute the material in any medium or format
--     Adapt:  remix, transform, and build upon the material

--     The licensor cannot revoke these freedoms as long as you follow
--     the license terms.

-- Under the following terms:

--     Attribution - You must give appropriate credit, provide a link to
--     the license, and indicate if changes were made. You may do so in
--     any reasonable manner, but not in any way that suggests the
--     licensor endorses you or your use.

--     NonCommercial - You may not use the material for commercial
--     purposes.

--     ShareAlike - If you remix, transform, or build upon the material,
--     you must distribute your contributions under the same license as
--     the original.

--     No additional restrictions - You may not apply legal terms or
--     technological measures that legally restrict others from doing
--     anything the license permits.
---------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

package util is
  --------------------------------------------------------------
  function b2x( b : boolean   ) return std_logic;
  function x2b( x : std_logic ) return boolean;
  --------------------------------------------------------------
  attribute dont_touch : boolean;
  --------------------------------------------------------------
  type std_logic_matrix is
    array( natural range <>, natural range <> ) of std_logic;
  --------------------------------------------------------------
  function to_downto( v : std_logic_vector ) return std_logic_vector;
  function rev ( v : std_logic_vector ) return std_logic_vector;
  --------------------------------------------------------------
  function lo( index, width : natural ) return natural;
  function hi( index, width : natural ) return natural;
  --------------------------------------------------------------
  function to_std_logic ( b : boolean ) return std_logic;
  function to_integer ( b : boolean ) return integer;
  --------------------------------------------------------------
  function replicate( b : std_logic; w : positive ) return std_logic_vector;
  function replicate_up( b : std_logic; w : positive ) return std_logic_vector;
  function pad_left( b : std_logic; v : std_logic_vector; w : natural )
    return std_logic_vector;
  function pad_right( v : std_logic_vector; b : std_logic; w : natural )
    return std_logic_vector;
  function pad_mod( char_width : natural; vector : std_logic_vector )
    return std_logic_vector;
  --------------------------------------------------------------
  function firstn ( v : std_logic_vector; w : natural ) return std_logic_vector;
  --------------------------------------------------------------
  function to_num_words( n, word_sz : natural ) return natural;
  function to_num_bits( n : natural ) return natural;
  function int_log( n, base : natural ) return natural;
  function int_log2( n : natural ) return natural;
  --------------------------------------------------------------
  function and_vec( v : std_logic_vector ) return std_logic;
  function or_vec( v : std_logic_vector ) return std_logic;
  function nor_vec( v : std_logic_vector ) return std_logic;
  function xor_vec( v : std_logic_vector ) return std_logic;
  --------------------------------------------------------------
end package;

package body util is

  --------------------------------------------------------------
  function b2x( b : boolean ) return std_logic is
  begin
    if b then
      return '1';
    else
      return '0';
    end if;
  end function;

  --------------------------------------------------------------
  function x2b( x : std_logic ) return boolean is
  begin
    if x='1' then
      return true;
    elsif x='0' then
      return false;
    else
      report( "warning: convert meta value to bool false");
      return false;
    end if;
  end function;

  --------------------------------------------------------------

  function rev( v : std_logic_vector ) return std_logic_vector is
    variable q_up : std_logic_vector( v'low to v'high );
    variable q_dn : std_logic_vector( v'high downto v'low );
  begin
    if v'ascending then
      for i in v'range loop
        q_up(i) := v( v'high - i + v'low );
      end loop;
      return q_up;
    else
      for i in v'range loop
        q_dn(i) := v( v'high - i + v'low );
      end loop;
      return q_dn;
    end if;
  end function;

  --------------------------------------------------------------

  function to_downto( v : std_logic_vector ) return std_logic_vector is
    variable z : std_logic_vector( v'high downto v'low );
  begin
    if v'ascending then
      -- report( "to_downto: ascending" );
      for i in v'range loop
        -- z(i) := v( v'high - i + v'low );
        z(i) := v( i );
      end loop;
      return z;
    else
      -- report( "to_downto: descending" );
      return v;
    end if;
  end function;

  --------------------------------------------------------------

  function lo( index, width : natural ) return natural is
  begin
    return index * width;
  end function;

  --------------------------------------------------------------

  function hi( index, width : natural ) return natural is
  begin
    return (index+1) * width - 1;
  end function;

  --------------------------------------------------------------

  function replicate( b : std_logic; w : positive ) return std_logic_vector is
    variable r : std_logic_vector( w-1 downto 0 ) := (others => b);
  begin
    return r;
  end function;

  function replicate_up( b : std_logic; w : positive ) return std_logic_vector is
    variable r : std_logic_vector( 0 to w-1 ) := (others => b);
  begin
    return r;
  end function;

  function pad_left( b : std_logic; v : std_logic_vector; w : natural )
    return std_logic_vector
  is
    variable n : integer;
  begin
    n := w - v'length;
    if n <= 0 then
      return v;
    else
      return replicate( b, n ) & v;
    end if;
  end;

  function pad_right( v : std_logic_vector; b : std_logic; w : natural )
    return std_logic_vector
  is
    variable n : integer;
  begin
    n := w - v'length;
    if n <= 0 then
      return v;
    else
      return v & replicate( b, n );
    end if;
  end;

  --------------------------------------------------------------

  function pad_mod( char_width : natural; vector : std_logic_vector )
    return std_logic_vector
  is
    variable n : natural;
  begin
    n := char_width - vector'length mod char_width;
    if n = char_width then
      n := 0;
    end if;
    -- report( "pad_mod: length="& natural'image(vector'length) &" pad="&
    --                            natural'image( n ) );
    return( std_logic_vector( to_unsigned(0, n) ) & vector );
  end function;

  --------------------------------------------------------------

  function firstn ( v : std_logic_vector; w : natural )
    return std_logic_vector
  is
  begin
    if v'ascending then
      return v( v'low to v'low + w - 1 );
    else
      return v( v'high downto v'high - w + 1 );
    end if;
  end function;

  --------------------------------------------------------------

  function to_std_logic ( b : boolean ) return std_logic is
  begin
    if b then return '1'; else return '0'; end if;
  end;

  --------------------------------------------------------------

  function to_integer ( b : boolean ) return integer is
  begin
    if b then return 1; else return 0; end if;
  end;

  --------------------------------------------------------------

  function to_num_words( n, word_sz : natural ) return natural is
    variable r : natural;
  begin
    r := n / word_sz;
    if n mod word_sz = 0 then
      return r;
    else
      return r+1;
    end if;
  end function;

  --------------------------------------------------------------

  function int_log( n, base : natural ) return natural is
  begin
    if n < base then
      return 1;
    else
      return (1 + int_log( n / base, base ));
    end if;
  end function;

  --------------------------------------------------------------

  function int_log2( n : natural ) return natural is
  begin
    return int_log( n, 2 );
  end function;

  --------------------------------------------------------------

  function to_num_bits( n : natural ) return natural is
    variable x : natural;
  begin
    x := int_log2( n );
    if n > 2 ** x then
      return x + 1;
    else
      return x;
    end if;
  end function;

  --------------------------------------------------------------
  -- return '1' if any of the bits of v are '1'
  --
  -- The advantage of this function over direct comparison to an array of
  -- '0' is that this function propagates 'X', just an OR does
  --
  function or_vec( v : std_logic_vector ) return std_logic is
    variable r : std_logic_vector( v'high downto v'low );
  begin
    r( r'low ) := v( v'low );
    for i in v'low+1 to v'high loop
      r(i) := v(i) or r(i-1);
    end loop;
    return r( r'high );
  end function;

  --------------------------------------------------------------
  -- return '1' if all of the bits of v are '0'

  function nor_vec( v : std_logic_vector ) return std_logic is
    variable r : std_logic_vector( v'high downto v'low );
  begin
    r( r'low ) := v( v'low );
    for i in v'low+1 to v'high loop
      r(i) := v(i) nor r(i-1);
    end loop;
    return r( r'high );
  end function;

  --------------------------------------------------------------
  -- return '1' if all of the bits of v are '1'
  --
  function and_vec( v : std_logic_vector ) return std_logic is
    variable r : std_logic_vector( v'high downto v'low );
  begin
    r( r'low ) := v( v'low );
    for i in v'low+1 to v'high loop
      r(i) := v(i) and r(i-1);
    end loop;
    return r( r'high );
  end function;

  --------------------------------------------------------------
  -- return '1' if an odd number of the bits of v are '1'
  --
  function xor_vec( v : std_logic_vector ) return std_logic is
    variable r : std_logic_vector( v'high downto v'low );
  begin
    r( r'low ) := v( v'low );
    for i in v'low+1 to v'high loop
      r(i) := v(i) xor r(i-1);
    end loop;
    return r( r'high );
  end function;

  --------------------------------------------------------------

end package body;
